@@ -50,9 +50,10 @@ public final class SSL {
     public static final int SSL_PROTOCOL_TLSV1 = (1<<2);
     public static final int SSL_PROTOCOL_TLSV1_1 = (1<<3);
     public static final int SSL_PROTOCOL_TLSV1_2 = (1<<4);
+    public static final int SSL_PROTOCOL_TLSV1_3 = (1<<5);
 
     /** TLS_*method according to <a href="https://www.openssl.org/docs/man1.0.2/ssl/SSL_CTX_new.html">SSL_CTX_new</a> */
-    public static final int SSL_PROTOCOL_TLS   = (SSL_PROTOCOL_SSLV3 | SSL_PROTOCOL_TLSV1 | SSL_PROTOCOL_TLSV1_1 | SSL_PROTOCOL_TLSV1_2);
+    public static final int SSL_PROTOCOL_TLS   = (SSL_PROTOCOL_SSLV3 | SSL_PROTOCOL_TLSV1 | SSL_PROTOCOL_TLSV1_1 | SSL_PROTOCOL_TLSV1_2 | SSL_PROTOCOL_TLSV1_3);
     public static final int SSL_PROTOCOL_ALL   = (SSL_PROTOCOL_SSLV2 | SSL_PROTOCOL_TLS);
 
     /*
@@ -69,6 +70,7 @@ public final class SSL {
     public static final int SSL_OP_NO_TLSv1 = sslOpNoTLSv1();
     public static final int SSL_OP_NO_TLSv1_1 = sslOpNoTLSv11();
     public static final int SSL_OP_NO_TLSv1_2 = sslOpNoTLSv12();
+    public static final int SSL_OP_NO_TLSv1_3 = sslOpNoTLSv13();
     public static final int SSL_OP_NO_TICKET = sslOpNoTicket();
 
     public static final int SSL_OP_NO_COMPRESSION = sslOpNoCompression();
@@ -500,10 +502,33 @@ public final class SSL {
      * @param ciphers an SSL cipher specification
      * @return {@code true} if successful
      * @throws Exception if an error happened
+     * @deprecated Use {@link #setCipherSuites(long, String, boolean)}
      */
-    public static native boolean setCipherSuites(long ssl, String ciphers)
-            throws Exception;
+    @Deprecated
+    public static boolean setCipherSuites(long ssl, String ciphers)
+            throws Exception {
+        return setCipherSuites(ssl, ciphers, false);
+    }
 
+    /**
+     * Returns the cipher suites available for negotiation in SSL handshake.
+     * <p>
+     * This complex directive uses a colon-separated cipher-spec string consisting
+     * of OpenSSL cipher specifications to configure the Cipher Suite the client
+     * is permitted to negotiate in the SSL handshake phase. Notice that this
+     * directive can be used both in per-server and per-directory context.
+     * In per-server context it applies to the standard SSL handshake when a
+     * connection is established. In per-directory context it forces a SSL
+     * renegotiation with the reconfigured Cipher Suite after the HTTP request
+     * was read but before the HTTP response is sent.
+     * @param ssl the SSL instance (SSL *)
+     * @param ciphers an SSL cipher specification
+     * @param tlsv13 {@code true} if the ciphers are for TLSv1.3
+     * @return {@code true} if successful
+     * @throws Exception if an error happened
+     */
+    public static native boolean setCipherSuites(long ssl, String ciphers, boolean tlsv13)
+            throws Exception;
     /**
      * Returns the ID of the session as byte array representation.
      *
@@ -531,7 +556,16 @@ public final class SSL {
      * @param ssl the SSL instance (SSL *)
      * @param hostname the hostname
      */
-    public static native void setTlsExtHostName(long ssl, String hostname);
+    public static void setTlsExtHostName(long ssl, String hostname) {
+        if (hostname != null && hostname.endsWith(".")) {
+            // Strip trailing dot if included.
+            // See https://github.com/netty/netty-tcnative/issues/400
+            hostname = hostname.substring(0, hostname.length() - 1);
+        }
+        setTlsExtHostName0(ssl, hostname);
+    }
+
+    private static native void setTlsExtHostName0(long ssl, String hostname);
 
     /**
      * Explicitly control <a href="https://wiki.openssl.org/index.php/Hostname_validation">hostname validation</a>
@@ -678,15 +712,31 @@ public final class SSL {
      * {@link #parseX509Chain(long)} and {@link #parsePrivateKey(long, String)}. It's important to note that the caller
      * of the method is responsible to free the passed in chain and key in any case as this method will increment the
      * reference count of the chain and key.
+     *
+     * @deprecated use {@link #setKeyMaterial(long, long, long)}
      */
-    public static native void setKeyMaterialServerSide(long ssl, long chain, long key) throws Exception;
+    @Deprecated
+    public static void setKeyMaterialServerSide(long ssl, long chain, long key) throws Exception {
+        setKeyMaterial(ssl, chain, key);
+    }
+
+    /**
+     * Sets the keymaterial to be used. The passed in chain and key needs to be generated via
+     * {@link #parseX509Chain(long)} and {@link #parsePrivateKey(long, String)}. It's important to note that the caller
+     * of the method is responsible to free the passed in chain and key in any case as this method will increment the
+     * reference count of the chain and key.
+     */
+    public static native void setKeyMaterial(long ssl, long chain, long key) throws Exception;
 
     /**
      * Sets the keymaterial to be used for the client side. The passed in chain and key needs to be generated via
      * {@link #parseX509Chain(long)} and {@link #parsePrivateKey(long, String)}. It's important to note that the caller
      * of the method is responsible to free the passed in chain and key in any case as this method will increment the
      * reference count of the chain and key.
+     *
+     * @deprecated use {@link #setKeyMaterial(long, long, long)}
      */
+    @Deprecated
     public static native void setKeyMaterialClientSide(long ssl, long x509Out, long pkeyOut, long chain, long key) throws Exception;
 
     /**
@@ -722,4 +772,21 @@ public final class SSL {
      * @throws Exception throws if setting the fips mode failed.
      */
     public static native void fipsModeSet(int mode) throws Exception;
+
+    /**
+     * Return the SNI hostname that was sent as part of the SSL Hello.
+     * @param ssl the SSL instance (SSL *)
+     * @return the SNI hostname or {@code null} if none was used.
+     */
+    public static native String getSniHostname(long ssl);
+
+    /**
+     * Return the signature algorithms that the remote peer supports or {@code null} if none are supported.
+     * See <a href="https://www.openssl.org/docs/man1.1.0/ssl/SSL_get_sigalgs.html"> man SSL_get_sigalgs</a> for more details.
+     * The returned names are generated using {@code OBJ_nid2ln} with the {@code psignhash} as parameter.
+     *
+     * @param ssl the SSL instance (SSL *)
+     * @return the signature algorithms or {@code null}.
+     */
+    public static native String[] getSigAlgs(long ssl);
 }
